/************************************************************************
 * Proiectarea cu Microprocesoare
 * Proiect 2013 - Hangman
 * Boulescu Adrian-Daniel
 ************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>

/************************************************************************ 
 * DEFINE-uri pentru parametrizarea modulului LCD
 ************************************************************************/
#define LcdDATA_DDR		DDRC			// Portul pe care conectam firele de date la LCD-ul
#define LcdDATA_PORT	PORTC
#define LcdDATA_PIN		PINC

#define LcdCMD_DDR		DDRC			// Portul pe care conectam firele de comenzi la LCD
#define LcdCMD_PORT		PORTC
#define LcdCMD_PIN		PINC

#define LcdD4			 PC3				// Pin-ul pentru firul de date D4 de pe LCD
#define LcdD5			 PC4				// Pin-ul pentru firul de date D5 de pe LCD
#define LcdD6			 PC5				// Pin-ul pentru firul de date D6 de pe LCD
#define LcdD7			 PC6				// Pin-ul pentru firul de date D7 de pe LCD

#define LcdRS			 PC0			    // Pinul pentru selectare operatie (LCD)
#define LcdRW			 PC1				// Pinul pentru Read/ Write (LCD)
#define LcdE			 PC2				// Pinul de Enable (LCD)

#define LCD_INSTR_4wire 		0x28 	// 4 fire date, font 5x8
#define LCD_INSTR_display 		0x0C 	// Display On, Cursor On, Blinking On ( 1 Display Cursor Blink )
#define LCD_INSTR_clearDisplay  	0x01 	// Clear Display
#define LCD_INSTR_returnHome		0x02 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_nextLine 		0xC0 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_gotoCGRAM		0x40	// go to Character Generator RAM

#define KEY_PORT PORTD           //Keypad Port
#define KEY_DDR DDRD         //Data Direction Register
#define KEY_PIN PIND         //Keypad Port Pins

#define COLUMN1 PD2
#define COLUMN2 PD0
#define COLUMN3 PD4
#define ROW1 PD1
#define ROW2 PD6
#define ROW3 PD5
#define ROW4 PD3

#define NCOLUMNS 3
#define NROWS 4

#define ASTERISK 20
#define DIEZ 30

#define NOT_PRESSED -1

/*************
 * API LCD   *
 *************/
void LCD_init();                // Initializare modul LCD.Trebuie apelata inainte de a se face orice operatie cu LCD-ul
void LCD_writeInstruction(unsigned char _instruction);	// Trimite o instructiune catre lcd (vezi datasheet)
void LCD_writeData(unsigned char _data);		// Trimite date catre LCD pentru afisare
void LCD_write(unsigned char _byte);			// trimite un bute catre LCD la modul general (nu conteaza daca e instructiune sau date)
void LCD_waitNotBusy();					// Asteptam pana cand lcd-ul devine disponibil pt o noua comanda
void LCD_waitInstructions(unsigned char _instructions);	// Asteapta un numar de cicli de ceas.
void LCD_print(char* _msg);				// Afiseaza imformatia pe LCD (doar 1 linie, primele 16 caractere)
void LCD_print2(char* _msg1, char* _msg2);		// Afisare pe 2 lini pe LCD
void LCD_printDecimal2u(unsigned int _n);		// Afisare numar in baza 10 pe LCD
void LCD_printHexa(unsigned int _n);			// Afisare numar in baza 16 pe LCD





/*****************************
 * IMPLEMENTARE API LCD.     *
 *****************************/

void LCD_init()
{
	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire

	LCD_waitNotBusy();
	
	LcdCMD_PORT   &= ~(1<<LcdRS);									// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);									// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD4)&~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |=  (1<<LcdD5);									// setam pinii de comenzi ca pini de iesire
	
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	LCD_waitInstructions(6);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );				// Setam linia E(nable) pe low; transferul s-a terminat

	LCD_writeInstruction(LCD_INSTR_4wire); 		// Incarcam comanda: 4 bit data, 2 lines, 5x8  font
	LCD_writeInstruction(LCD_INSTR_display); 	// Display On, Cursor On, Blinking On
	LCD_writeInstruction(0x06);					// Increment, no shift
	LCD_writeInstruction(0x01);					// Clear Display

	
}

void LCD_writeInstruction(unsigned char _instruction)
{				
	LCD_waitNotBusy();							// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LcdDATA_PORT &= ~(1 << LcdRS);				// setam pinul RS pe low (low=instructiuni, high=date)
	LcdDATA_PORT &= ~(1 << LcdRW);				// setam pinul RW pe low (suntem in modul de comenzi acum)
	LCD_write(_instruction);					// apelam procedura ce trimite byte-ul pe firele de date
}


void LCD_writeData(unsigned char _data)
{				
	LCD_waitNotBusy();							// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LcdDATA_PORT |= (1 << LcdRS);				// setam pinul RS pe low (low=instructiuni, high=date)
	LcdDATA_PORT &= ~(1 << LcdRW);				// setam pinul RW pe low (suntem in modul de comenzi acum)
	LCD_write(_data);					// apelam procedura ce trimite byte-ul pe firele de date
}

void LCD_write(unsigned char _byte)
{
	// scriem cei mai semnificativi 4 biti pe firele de date
	if ((_byte & (1 << 7)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD7);
	else
		LcdDATA_PORT |= (1 << LcdD7);
		
	if ((_byte & (1 << 6)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD6);
	else
		LcdDATA_PORT |= (1 << LcdD6);
		
	if ((_byte & (1 << 5)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD5);
	else
		LcdDATA_PORT |= (1 << LcdD5);
		
	if ((_byte & (1 << 4)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD4);
	else
		LcdDATA_PORT |= (1 << LcdD4);

	LcdDATA_PORT |= (1 << LcdE);	// Setam Pinul E pe high
	LCD_waitInstructions(6);		// Asteptam o perioada de timp T
	LcdDATA_PORT &= ~(1 << LcdE);	// Setam Pinul E pe low
	LCD_waitInstructions(6);		// Asteptam o perioada de timp T
	
	// scriem cei mai putin semnificativi 4 biti pe firele de date
	if ((_byte & (1 << 3)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD7);
	else
		LcdDATA_PORT |= (1 << LcdD7);
		
	if ((_byte & (1 << 2)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD6);
	else
		LcdDATA_PORT |= (1 << LcdD6);
		
	if ((_byte & (1 << 1)) == 0)
		LcdDATA_PORT &= ~(1 << LcdD5);
	else
		LcdDATA_PORT |= (1 << LcdD5);
		
	if ((_byte & 1) == 0)
		LcdDATA_PORT &= ~(1 << LcdD4);
	else
		LcdDATA_PORT |= (1 << LcdD4);

	LcdDATA_PORT |= (1 << LcdE);	// Setam Pinul E pe high
	LCD_waitInstructions(6);		// Asteptam o perioada de timp T
	LcdDATA_PORT &= ~(1 << LcdE);	// Setam Pinul E pe low
}

void LCD_waitNotBusy()
{
	unsigned char _loop = 1;
	_delay_ms(50);
	while (_loop)
	{
		LcdDATA_DDR &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7);	// Setam pinii de date de la LCD pe in pt a citi busy flag
		LcdDATA_PORT &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); 	// Dezactivam pullup resistor pentru pinii de in
		
		LcdCMD_PORT &= ~(1<<LcdE);					// Setam pin-ul e pe low; ar trebui sa fie deja low, doar ne asiguram
		LcdCMD_PORT &= ~(1<<LcdRS);					// Setam pinul RS pe low
		LcdCMD_PORT |=  (1<<LcdRW);					// Setam pinul RW pe high (acum suntem in modul de interogare busy/adr)
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_loop = LcdDATA_PIN & (1<<LcdD7);				// Citim busy flag-ul
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
		
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		LcdDATA_DDR |= (1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); // Setam Portul de LCD ca port de iesire la loc
	}
}

void LCD_waitInstructions(unsigned char _instructions)
{
	while (_instructions--)
		;
}


void LCD_printDecimal2u(unsigned int _n)
{
	unsigned char tmp=0;
	
	// Extragem sutele
	while(_n>=100)
		_n-=100;

	while(_n>=10){
		tmp++;
		_n-=10;
	}

	LCD_writeData(tmp+'0');
	LCD_writeData(_n+'0');
}

void LCD_printHexa(unsigned int _n)
{
	unsigned char _tmp = _n>>4;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
	_tmp = _n & 0x0F;
	if (_tmp>9)
		_tmp += 'A'-10;
	else
		_tmp += '0';
	LCD_writeData( _tmp );
}


void LCD_print(char* _msg)
{
	unsigned char i=0;
	for( ; _msg[i]!=0 && i<64; i++)
		LCD_writeData( _msg[i] );
}

void LCD_printn(char* _msg, uint8_t n)
{
	unsigned char i=0;
	for( ; _msg[i]!=0 && i<n; i++)
		LCD_writeData( _msg[i] );
}

void LCD_print2(char* _msg1, char* _msg2)
{
	LCD_writeInstruction(LCD_INSTR_clearDisplay);
	LCD_print(_msg1);
	LCD_writeInstruction(LCD_INSTR_nextLine);
	LCD_print(_msg2);
}

/*****************************
 * IMPLEMENTARE API KEYPAD.  *
 *****************************/

int getkey()
{
   PORTD |= (1 << ROW2) | (1 << ROW1) | (1 << ROW3) | (1 << ROW4);  // activam rezistenta de pull-up
  
   DDRD |= (1 << COL1) | (1 << COL2) | (1 << COL3);
   int i, ret = -1;
   
   for (i = 0; i < 3; i++) {
      switch(i) {
	 case 0: {
	  DDRD |= (1 << COL1);
	  
	  if (!(PIND & (1 << ROW1)))
	     ret = 1;
	  if (!(PIND & (1 << ROW2)))
	     ret = 4
	  if (!(PIND & (1 << ROW3)))
	     ret = 7;
	  if (!(PIND & (1 << ROW4)))
	     ret = ASTERISK;
	  
	  DDRD &= ~(1<< COL1);
	  return ret;
	 }
	 case 1: {
	  DDRD |= 1 << COL2;  
	  
	  if (!(PIND & (1 << ROW1)))
	     ret = 2;
	  if (!(PIND & (1 << ROW2)))
	     ret = 5;
	  if (!(PIND & (1 << ROW3)))
	     ret = 8;
	  if (!(PIND & (1 << ROW4)))
	     ret = 0;
	  
	  DDRD &= ~(1 << COL2);
	  return ret;
	 }
	 case 2: {
	  DDRD |= 1 << COL3;  
	  
	  if (!(PIND & (1 << ROW1)))
	     ret = 3;
	  if (!(PIND & (1 << ROW2)))
	     ret = 6;
	  if (!(PIND & (1 << ROW3)))
	     ret = 9;
	  if (!(PIND & (1 << ROW4)))
	     ret = DIEZ;
	  
	  DDRD &= ~(1<< COL3);
	  return ret;
	 }
      }
   }
   return ret;
}


int valid(char *word, char letter, char **toguess)
{
	int i, found, guessed;
	
	found = 0;
	
	for (i = 0; i < strlen(word); ++i) {
		if (word[i] == letter) {
			(*toguess)[i] = letter;
			found = 1;
		}
	}
	
	guessed = 1;
	for (i = 0; i < strlen(word); ++i) {
		if ((*toguess)[i] == '_') {
			guessed = 0;
		}
	}
	
	if (guessed == 1) {
		return 2;
	}
	
	return found;
}

int main()
{
   int p, lastpressed, count, r, x, chosen, response, tries, i;
	char letters[8][3] = {"abc", "def", "ghi", "jkl", "mno", "prs", "tuv", 
																"wxy"};
	
	char words[9][16] = {"nedeterminat", "revedere", "altruist", "itinerariu", "toluen", 
							"lingvistic", "virtuos", "cognitiv", "anticipare"};
	char *guessed = malloc(16 * sizeof(char));
	char msg[16];

	LCD_init();
	
	KEY_DDR = 0xF0;
	KEY_PORT = 0x00;
	
	DDRC = 0xFF;
	PORTC = 0xFF;
	PIND = 0;

	count = 0;
	chosen = 0;
	tries = 0;
	lastpressed = NOT_PRESSED;
	
	memset(guessed, 0, 16*sizeof(char));
	
	LCD_print2("Press # to start:", "DIEZ");
	
	while (1) {
		p = get_key();	// get pressed key
		if (p != NOT_PRESSED) {
			if (p == 30 && lastpressed != NOT_PRESSED) {		// '#'
				if (count == 0 || count > 3) {
					count = 0;
					continue;
				}
				r = lastpressed - (lastpressed / 4);
				
				if (chosen == 0) {
					chosen = 1;
					x = r;
					for (i = 0; i < strlen(words[x]); ++i) {
						guessed[i] = '_';
					}
					valid(words[x], words[x][0], &guessed);
					LCD_writeInstruction(LCD_INSTR_clearDisplay);
					sprintf(msg, "%s", guessed);
					LCD_print(msg);
				}
				else {
					response = valid(words[x], letters[r][count-1], &guessed);
					
					if (response == 2) {
						LCD_writeInstruction(LCD_INSTR_clearDisplay);
						sprintf(msg, "%s", guessed);
						LCD_print2("You won!", msg);
						break;
					}
					
					if (response == 0) {
						tries++;
					}
					
					
					if (tries == 5) {
						LCD_writeInstruction(LCD_INSTR_clearDisplay);
						LCD_print2("Game over", " :( ");
						break;
					}
					
					LCD_writeInstruction(LCD_INSTR_clearDisplay);
					sprintf(msg, "%s", guessed);
					LCD_print(msg);
				}
				count = 0;
				continue;
			}
				
			if (p != lastpressed) {
				count = 1;
				lastpressed = p;
			}
			else {
				count++;
			}
		}
	}
	
	free(guessed);
}
